'''
This code is lightly adapted from https://github.com/pucrs-automated-planning/pddl-parser
'''

import itertools
import re
import sys
import pprint

from bddl.config import SUPPORTED_BDDL_REQUIREMENTS as supported_requirements
from bddl.config import get_domain_filename, get_definition_filename, READABLE_PREDICATE_NAMES


def scan_tokens(filename=None, string=None):
    if filename is not None:
        with open(filename, 'r') as f:
            # Remove single line comments
            raw_str = f.read()
    elif string is not None:
        raw_str = string
    else:
        raise ValueError("No input BDDL provided.")
    str = re.sub(r';.*$', '', raw_str, flags=re.MULTILINE).lower()
    # Tokenize
    stack = []
    tokens = []
    for t in re.findall(r'[()]|[^\s()]+', str):
        if t == '(':
            stack.append(tokens)
            tokens = []
        elif t == ')':
            if stack:
                toks = tokens
                tokens = stack.pop()
                tokens.append(toks)
            else:
                raise Exception('Missing open parenthesis')
        else:
            tokens.append(t)
    if stack:
        raise Exception('Missing close parenthesis')
    if len(tokens) != 1:
        raise Exception('Malformed expression')
    return tokens[0]


def parse_domain(domain):
    domain_filename = get_domain_filename(domain)
    tokens = scan_tokens(filename=domain_filename)
    if type(tokens) is list and tokens.pop(0) == 'define':
        domain_name = 'unknown'
        requirements = []
        types = []
        actions = []
        predicates = {}
        while tokens:
            group = tokens.pop(0)
            t = group.pop(0)
            if t == 'domain':
                domain_name = group[0]
            elif t == ':requirements':
                for req in group:
                    if not req in supported_requirements:
                        raise Exception('Requirement %s not supported' % req)
                requirements = group
            elif t == ':predicates':
                # predicate_name, arguments = parse_predicates(group)
                predicates = parse_predicates(group)
            elif t == ':types':
                types = group
            elif t == ':action':
                name = group.pop(0)
                for act in actions:
                    if act.name == name:
                        raise Exception(
                            'Action %s is defined multiple times' % name)
                actions.append(parse_action(group))
            else:
                print('%s is not recognized in domain' % t)
        return domain_name, requirements, types, actions, predicates
    else:
        raise Exception('File %s does not match domain pattern' %
                        domain_filename)


def parse_predicates(group):
    predicates = {}
    for pred in group:
        predicate_name = pred.pop(0)
        arguments = {}
        untyped_variables = []
        while pred:
            t = pred.pop(0)
            if t == '-':
                if not untyped_variables:
                    raise Exception('Unexpected hyphen in predicates')
                var_type = pred.pop(0)
                while untyped_variables:
                    arguments[untyped_variables.pop(0)] = var_type
            else:
                untyped_variables.append(t)
        while untyped_variables:
            arguments[untyped_variables.pop(0)] = 'object'
        # print(predicates)
        predicates[predicate_name] = arguments
    return predicates


def parse_action(group):
    name = group.pop(0)
    if not isinstance(name, str):
        raise Exception('Action without name definition')
    parameters = []
    positive_preconditions = []
    negative_preconditions = []
    add_effects = []
    del_effects = []
    while group:
        t = group.pop(0)
        if t == ':parameters':
            if not isinstance(group, list):
                raise Exception('Error with %s parameters' % name)
            parameters = []
            untyped_parameters = []
            p = group.pop(0)
            while p:
                t = p.pop(0)
                if t == '-':
                    if not untyped_parameters:
                        raise Exception(
                            'Unexpected hyphen in %s parameters' % name)
                    param_type = p.pop(0)
                    while untyped_parameters:
                        parameters.append(
                            [untyped_parameters.pop(0), param_type])
                else:
                    untyped_parameters.append(t)
            while untyped_parameters:
                parameters.append([untyped_parameters.pop(0), 'object'])
        elif t == ':precondition':
            split_predicates(group.pop(0), positive_preconditions,
                             negative_preconditions, name, ' preconditions')
        elif t == ':effect':
            split_predicates(group.pop(0), add_effects,
                             del_effects, name, ' effects')
        else:
            print('%s is not recognized in action' % t)
    return Action(name, parameters, positive_preconditions, negative_preconditions, add_effects, del_effects)


def parse_problem(behavior_activity, activity_definition, domain_name, predefined_problem=None):
    if predefined_problem is not None:
        tokens = scan_tokens(string=predefined_problem)
    else:
        problem_filename = get_definition_filename(
            behavior_activity, activity_definition)
        tokens = scan_tokens(filename=problem_filename)
    if isinstance(tokens, list) and tokens.pop(0) == 'define':
        problem_name = 'unknown'
        objects = {}
        initial_state = []
        goal_state = []
        while tokens:
            group = tokens.pop()
            t = group[0]
            if t == 'problem':
                problem_name = group[-1]
            elif t == ':domain':
                if domain_name != group[-1]:
                    raise Exception(
                        'Different domain specified in problem file')
            elif t == ':requirements':
                pass
            elif t == ':objects':
                group.pop(0)
                object_list = []
                while group:
                    if group[0] == '-':
                        group.pop(0)
                        objects[group.pop(0)] = object_list
                        object_list = []
                    else:
                        object_list.append(group.pop(0))
                if object_list:
                    if not 'object' in objects:
                        objects['object'] = []
                    objects['object'] += object_list
            elif t == ':init':
                group.pop(0)
                initial_state = group
            elif t == ':goal':
                package_predicates(group[1], goal_state, '', 'goals')
            else:
                print('%s is not recognized in problem' % t)
        return problem_name, objects, initial_state, goal_state
    else:
        raise Exception(
            f"Problem {behavior_activity} {activity_definition} does not match problem pattern")


def split_predicates(group, pos, neg, name, part):
    if not isinstance(group, list):
        raise Exception('Error with ' + name + part)
    if group[0] == 'and':
        group.pop(0)
    else:
        group = [group]
    for predicate in group:
        if predicate[0] == 'not':
            if len(predicate) != 2:
                raise Exception('Unexpected not in ' + name + part)
            # NOTE removed this because I want the negative goals to have "not"
            neg.append(predicate[-1])
        else:
            pos.append(predicate)


def package_predicates(group, goals, name, part):
    if not isinstance(group, list):
        raise Exception('Error with ' + name + part)
    if group[0] == 'and':
        group.pop(0)
    else:
        group = [group]
    for predicate in group:
        goals.append(predicate)


class Action(object):

    def __init__(self, name, parameters, positive_preconditions, negative_preconditions, add_effects, del_effects):
        self.name = name
        self.parameters = parameters
        self.positive_preconditions = positive_preconditions
        self.negative_preconditions = negative_preconditions
        self.add_effects = add_effects
        self.del_effects = del_effects

    def __str__(self):
        return 'action: ' + self.name + \
            '\n  parameters: ' + str(self.parameters) + \
            '\n  positive_preconditions: ' + str(self.positive_preconditions) + \
            '\n  negative_preconditions: ' + str(self.negative_preconditions) + \
            '\n  add_effects: ' + str(self.add_effects) + \
            '\n  del_effects: ' + str(self.del_effects) + '\n'

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def groundify(self, objects):
        if not self.parameters:
            yield self
            return
        type_map = []
        variables = []
        for var, type in self.parameters:
            type_map.append(objects[type])
            variables.append(var)
        for assignment in itertools.product(*type_map):
            positive_preconditions = self.replace(
                self.positive_preconditions, variables, assignment)
            negative_preconditions = self.replace(
                self.negative_preconditions, variables, assignment)
            add_effects = self.replace(self.add_effects, variables, assignment)
            del_effects = self.replace(self.del_effects, variables, assignment)
            yield Action(self.name, assignment, positive_preconditions, negative_preconditions, add_effects, del_effects)

    def replace(self, group, variables, assignment):
        g = []
        for pred in group:
            pred = list(pred)
            iv = 0
            for v in variables:
                while v in pred:
                    pred[pred.index(v)] = assignment[iv]
                iv += 1
            g.append(pred)
        return g


######### WRITING UTILS ##########

def flatten_list(li):
    for elem in li:
        if isinstance(elem, list):
            yield from flatten_list(elem)
        else:
            yield elem


def gen_natural_language_condition(parsed_condition, indent=0):
    indent_string = " " * 4 * indent
    # print(parsed_condition)
    term = parsed_condition

    # for term in parsed_condition:
    if isinstance(term, list):
        if any([isinstance(subterm, list) for subterm in term]):
            if term[0] == "and":
                print('indent from and:', indent)
                yield f", and\n".join([list(gen_natural_language_condition(subterm, indent=indent + 1))[0] for subterm in term[1:]])
            elif term[0] == "or":
                yield f", or\n".join([list(gen_natural_language_condition(subterm, indent=indent + 1))[0] for subterm in term[1:]]) + \
                    f",\n{indent_string + '    '}or any combination of these"
            elif term[0] == "not":
                yield indent_string + "the following is NOT true:\n" + list(gen_natural_language_condition(term[1], indent=indent + 1))[0]
            elif term[0] == "imply":
                yield indent_string + f"if\n{list(gen_natural_language_condition(term[1], indent=indent + 1))[0]}\n{indent_string}then\n{list(gen_natural_language_condition(term[2], indent=indent + 1))[0]}\n{indent_string}but if not it doesn't matter"
            elif term[0] == "forall":
                yield indent_string + f"for every {nlterm(term[1][0])},\n{list(gen_natural_language_condition(term[2], indent=indent + 1))[0]}"
            elif term[0] == "exists":
                yield indent_string + f"for at least one {nlterm(term[1][0])},\n{list(gen_natural_language_condition(term[2], indent=indent + 1))[0]}"
            elif term[0] == "forn":
                yield indent_string + f"for exactly {term[1][0]} {nlterm(term[2][0])}(s),\n{list(gen_natural_language_condition(term[3], indent=indent + 1))[0]}"
            elif term[0] == "forpairs":
                yield indent_string + f"for pairs of {nlterm(term[1][0])}s and {nlterm(term[2][0])}s,\n{list(gen_natural_language_condition(term[3], indent=indent + 1))[0]}"
            elif term[0] == "fornpairs":
                yield indent_string + f"for exactly {term[1][0]} pairs of {nlterm(term[2][0])}s and {nlterm(term[3][0])}s,\n{list(gen_natural_language_condition(term[4], indent=indent + 1))[0]}"

        else:
            # print(indent)
            if len(term) == 2:
                # fixed_term1 = term[1].lstrip('?').split('.')[0]
                # if '_' in term[1]:
                #     fixed_term1 += term[1].split('_')[-1]
                article1 = "the " if "_" not in term[1] else ""
                desc = READABLE_PREDICATE_NAMES[term[0]
                                                ] if term[0] in READABLE_PREDICATE_NAMES else term[0]
                yield f"{indent_string}{article1}{nlterm(term[1])} is {desc}"
            elif len(term) == 3:
                article1 = "the " if "_" not in term[1] else ""
                article2 = "the " if "_" not in term[2] else ""
                desc = READABLE_PREDICATE_NAMES[term[0]
                                                ] if term[0] in READABLE_PREDICATE_NAMES else term[0]
                yield f"{indent_string}{article1}{nlterm(term[1])} is {desc} {article2}{nlterm(term[2])}"

    else:
        raise ValueError('encountered non-list:', term)
        yield ''


def nlterm(term):
    natural_term = term.lstrip('?')
    natural_term = natural_term.split('.')[0]
    if '_' in term:
        # print('it is present')
        natural_term += term.split('_')[-1]
    return natural_term


def gen_natural_language_conditions(parsed_conditions):
    return [''.join(list(gen_natural_language_condition(parsed_condition))) for parsed_condition in parsed_conditions]


def add_bddl_whitespace(bddl_file="activity_conditions/parsing_tests/test_app_output.bddl", string=None, save=True):
    if string is not None:
        raw_bddl = string
    elif bddl_file is not None:
        with open(bddl_file, "r") as f:
            raw_bddl = f.read()
    else:
        raise ValueError("No BDDL given")

    total_characters = len(raw_bddl)

    nest_level = 0
    refined_bddl = ""
    new_block = ""
    char_i = 0
    last_paren_type = None
    while char_i < total_characters:
        if raw_bddl[char_i] == "(":
            new_block = '\n' + '    ' * nest_level + raw_bddl[char_i]
            last_paren_type = "("
            char_i += 1
            while (raw_bddl[char_i] not in [' ', ')']) and char_i < total_characters:
                new_block += raw_bddl[char_i]
                char_i += 1
            refined_bddl += new_block + raw_bddl[char_i]
            if raw_bddl[char_i] == ' ':
                nest_level += 1
        elif raw_bddl[char_i] == ")":
            nest_level -= 1
            if last_paren_type == ")":
                refined_bddl += "\n" + '    ' * nest_level
            refined_bddl += raw_bddl[char_i]
            last_paren_type = ")"
        else:
            refined_bddl += raw_bddl[char_i]
        char_i += 1

    if save:
        with open('activity_definitions/parsing_tests/test_app_output_whitespace.bddl', 'w') as f:
            f.write(refined_bddl)

    return refined_bddl


def remove_bddl_whitespace(
    bddl_file='activity_definitions/parsing_tests/test_app_output_whitespace.bddl', 
    string=None, 
    save=True,
    outfile='activity_definitions/parsing_tests/test_app_output_nowhitespace.bddl'
):    
    if bddl_file is not None:
        with open(bddl_file, 'r') as f:
            raw_bddl = f.read()
    elif string is not None:
        raw_bddl = string
    else:
        raise ValueError('No BDDL given.')

    bddl = ' '.join([substr.lstrip(' ') for substr in raw_bddl.split('\n')])
    bddl = [' ' + substr if substr[0] !=
            ')' else substr for substr in bddl.split(' ') if substr]
    bddl = ''.join(bddl)[1:]

    if save:
        with open(outfile, 'w') as f:
            f.write(bddl)

    return bddl


def construct_full_bddl(behavior_activity, activity_definition, object_list, init_state, goal_state):
    """Make full BDDL problem file from parts, release as string

    :param object_list (string): object list (assumed whitespace added with tabs)   TODO change assumptions if needed
    :param init_state (string): initial state (assumed whitespace not added)
    :param goal_state (string): goal state (assumed whitespace not added)
    """
    object_list = "    ".join(object_list.split("\t"))
    init_state = "    \n".join(add_bddl_whitespace(
        bddl_file=None, string=init_state, save=False).split("\n"))
    goal_state = "    \n".join(add_bddl_whitespace(
        bddl_file=None, string=goal_state, save=False).split("\n"))
    bddl = f"""(define\n
                   (problem {behavior_activity}_{activity_definition})\n
                   (:domain igibson)\n
                {object_list}\n
                {init_state}\n
                {goal_state}\n
               )"""
    return bddl


def construct_bddl_from_parsed(behavior_activity, activity_definition, parsed_object_list, parsed_init_state, parsed_goal_state, domain="omnigibson"):
    object_list = "(:objects\n"
    for object_cat, object_insts in parsed_object_list.items():
        object_list += f"        " + " ".join(object_insts) + f" - {object_cat}\n"
    object_list += "    )"

    init_state = f"(:init"
    for literal in parsed_init_state:
        if literal[0] == "not": 
            init_state += f" (not ({' '.join(literal[1])}))"
        else:
            init_state += f" ({' '.join(literal)})"
    init_state += ")"
    init_state = add_bddl_whitespace(string=init_state, save=False)
    init_state = "\n    ".join(init_state.split("\n"))
    
    goal_state = f"(:goal {build_goal(['and'] + parsed_goal_state)})"
    goal_state = add_bddl_whitespace(string=goal_state, save=False)
    goal_state = "\n    ".join(goal_state.split("\n"))

    bddl = f"""(define (problem {behavior_activity}_{activity_definition})
    (:domain {domain})\n
    {object_list}
    {init_state}
    {goal_state}
)"""
    return bddl


def build_goal(goal_expr):
    if type(goal_expr[1]) == list:
        return f"({goal_expr[0]} {' '.join([build_goal(subexpr) for subexpr in goal_expr[1:]])})"
    else: 
        return f"({' '.join(goal_expr)})"


if __name__ == '__main__':
    if sys.argv[1] == 'add':
        refined_bddl = add_bddl_whitespace()
    if sys.argv[1] == 'remove':
        refined_bddl = remove_bddl_whitespace()
    if sys.argv[1] == "construct_from_parsed":
        activity = "cleaning_up_after_a_meal"
        __, objs, init, goal = parse_problem(activity, 0, "igibson")
        reconstruction = construct_bddl_from_parsed(activity, 0, objs, init, goal, domain="igibson")
        with open(f"activity_definitions/{activity}/problem0.bddl", "r") as f:
            defn_lines = f.read().split("\n")
        reconstruction_lines = reconstruction.split("\n")
        print("Length same:", len(defn_lines) == len(reconstruction_lines))
        print(len(defn_lines))
        print(len(reconstruction_lines))
        # pprint.pprint(defn_lines)
        # pprint.pprint(reconstruction_lines)
        # print(reconstruction)
        for i in range(len(reconstruction_lines)): 
            # print(defn_lines[i].strip() == reconstruction_lines[i].strip())
            if defn_lines[i].strip() != reconstruction_lines[i].strip():
                print(defn_lines[i].strip())
                print(reconstruction_lines[i].strip())
                print(f"{i} doesn't match")
                print()
                import sys; sys.exit()
            else:
                continue
                # print(i)



import csv
from enum import IntEnum
import json
import os
import re
import pandas as pd
import pathlib


OBJECT_INSTANCE_RE = r"[A-Za-z-_]+\.n\.[0-9]+_[0-9]+"
OBJECT_CAT_RE = r"[A-Za-z-_]+\.n\.[0-9]+$"
OBJECT_CAT_AND_INST_RE = r"[A-Za-z-_]+\.n\.[0-9]+"

SYNS_TO_PROPS = pathlib.Path(__file__).parents[1] / "generated_data" / "propagated_annots_canonical.json"
PROP_PARAM_ANNOTS_DIR = pathlib.Path(__file__).parents[1] / "generated_data" / "prop_param_annots"
PARAMS_OUTFILE_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "propagated_annots_params.json"
REMOVER_SYNSET_MAPPING = pathlib.Path(__file__).parents[1] / "generated_data" / "remover_synset_mapping.json"
LEAF_SYNSETS_FILE = pathlib.Path(__file__).parents[1] / "generated_data" / "synsets.csv"

# Helpers
LEAF_SYNSETS_SET = set(pd.read_csv(LEAF_SYNSETS_FILE)["synset"])

# Specific condition types for applying / removing particles
class ParticleModifyCondition(IntEnum):
    FUNCTION = 0
    SATURATED = 1
    TOGGLEDON = 2
    GRAVITY = 3

PREDICATE_MAPPING = {
    "saturated": ParticleModifyCondition.SATURATED,
    "toggled_on": ParticleModifyCondition.TOGGLEDON,
    "function": ParticleModifyCondition.FUNCTION,
}

def parse_predicate(condition):
    predicate = condition.split(" ")[0]
    pred_type = PREDICATE_MAPPING[predicate]
    if pred_type == ParticleModifyCondition.SATURATED:
        cond = (predicate, condition.split(" ")[1])
    elif pred_type == ParticleModifyCondition.TOGGLEDON:
        cond = (predicate, True)
    elif pred_type == ParticleModifyCondition.FUNCTION:
        raise ValueError("Not supported")
    else:
        raise ValueError(f"Unsupported condition type: {pred_type}")
    return cond


def parse_conditions_entry(unparsed_conditions):
    # print(f"Parsing: {unparsed_conditions}")
    if unparsed_conditions.isnumeric():
        always_true = bool(int(unparsed_conditions))
        conditions = [] if always_true else None
    else:
        conditions = [parse_predicate(condition=cond) for cond in unparsed_conditions.lower().split(" or ")]
    return conditions



# particleRemovers

def get_synsets_to_particle_remover_params():
    synset_cleaning_mapping = {}
    records = []
    with open(os.path.join(PROP_PARAM_ANNOTS_DIR, "particleRemover.csv"), "r") as f:
        reader = csv.DictReader(f)
        for row in reader:
            records.append(row)

    for record in records: 
        synset = re.match(OBJECT_CAT_AND_INST_RE, record["synset"])
        assert synset is not None, "prop_param_annots/particleRemover.csv has line without properly formed synset"
        synset = synset.group()

        # Skip washer and dryer
        if "not particleremover" in record["synset"].lower(): continue

        default_fluid_conditions = parse_conditions_entry(record["other liquids"])
        default_visual_conditions = parse_conditions_entry(record["other visualSubstances"])
        default_non_fluid_conditions = parse_conditions_entry(record["other physicalSubstances"])
        if record["method"] not in {"projection", "adjacency"}:
            raise ValueError(f"Synset {record['synset']} prop particleRemover has invalid method {record['method']}")
        
        remover_kwargs = {
            "conditions": {},
            "default_visual_conditions": default_visual_conditions,
            "default_non_fluid_conditions": default_non_fluid_conditions,
            "default_fluid_conditions": default_fluid_conditions,
            "method": record["method"],
        }

        # Iterate through all the columns headed by a substance, in no particular order since their ultimate location is a dict
        for dirtiness_substance_synset in [key for key in record if re.match(OBJECT_CAT_AND_INST_RE, key) is not None]:
            conditions = parse_conditions_entry(record[dirtiness_substance_synset])
            remover_kwargs["conditions"][dirtiness_substance_synset] = conditions

        synset_cleaning_mapping[synset] = remover_kwargs
    
    return synset_cleaning_mapping


# Main parameter method

def create_get_save_propagated_annots_params(syns_to_props):
    print("Processing param annots...")
    synsets_to_particleremover_params = get_synsets_to_particle_remover_params()
    for particleremover_syn, particleremover_params in synsets_to_particleremover_params.items():
        assert "particleRemover" in syns_to_props[particleremover_syn], f"Synset {particleremover_syn} has particleRemover params but is not annotated as a particleRemover"
        syns_to_props[particleremover_syn]["particleRemover"] = particleremover_params

    for prop_fn in os.listdir(PROP_PARAM_ANNOTS_DIR):
        prop = prop_fn.split(".")[0]

        if prop == "particleRemover": continue
        else:
            param_annots = pd.read_csv(os.path.join(PROP_PARAM_ANNOTS_DIR, prop_fn)).to_dict(orient="records")
            for param_record in param_annots: 
                for param_name, param_value in param_record.items():

                    if param_name == "synset": continue

                    if param_name == "requires_toggled_on" and param_value == 1:
                        assert "toggleable" in syns_to_props[param_record["synset"]], f"Synset {param_record['synset']} has requires_toggled_on param but is not annotated as toggleable"
                    if param_name == "requires_closed" and param_value == 1:
                        assert "openable" in syns_to_props[param_record["synset"]], f"Synset {param_record['synset']} has requires_closed param but is not annotated as openable"

                    # Params that can have NaN values
                    if param_name in [
                        "substance_cooking_derivative_synset", 
                        "sliceable_derivative_synset", 
                        "uncooked_diceable_derivative_synset",
                        "cooked_diceable_derivative_synset"
                    ]:
                        if not pd.isna(param_value):
                            formatted_param_value = param_value
                            syns_to_props[param_record["synset"]][prop][param_name] = formatted_param_value
                        continue

                    # Params that should not have NaN values

                    # NaN values
                    elif pd.isna(param_value):
                        if prop == "coldSource":
                            raise ValueError(f"synset {param_record['synset']} coldSource annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "cookable":
                            raise ValueError(f"synset {param_record['synset']} cookable annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "flammable":
                            raise ValueError(f"synset {param_record['synset']} flammable annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "heatable":
                            raise ValueError(f"synset {param_record['synset']} heatable annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "heatSource":
                            raise ValueError(f"synset {param_record['synset']} heatSource annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "particleApplier":
                            raise ValueError(f"synset {param_record['synset']} particleApplier annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "particleSink":
                            if param_name == "conditions": 
                                formatted_param_value = {}
                            elif param_name == "default_fluid_conditions":
                                formatted_param_value = []
                            elif param_name == "default_non_fluid_conditions":
                                formatted_param_value = []
                            elif param_name == "default_visual_conditions":
                                formatted_param_value = None
                            else:
                                raise ValueError(f"synset {param_record['synset']} particleSink annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "particleSource":
                            raise ValueError(f"synset {param_record['synset']} particleSource annotation has NaN value for parameter {param_name}. Either handle NaN or annotate parameter value.")
                        elif prop == "meltable": 
                            raise ValueError(f"synset {param_record['synset']} meltable annotation has NaN value for parameter {param_name}. Either handle NaN or annotate param value.")
                    
                    
                    # `conditions` values - format is keyword1:bool_value1;keyword2:bool_value2;...;keywordN:bool_valueN
                    elif param_name == "conditions": 
                        if (prop == "particleApplier") or (prop == "particleSource"):
                            # NOTE assumes that there may be multiple conditions
                            conditions = param_value.split(";")
                            formatted_conditions = []
                            for condition in conditions: 
                                if condition == "gravity:True":
                                    formatted_conditions.append(("gravity", True))
                                elif condition == "toggled_on:True":
                                    assert "toggleable" in syns_to_props[param_record["synset"]], f"Synset {param_record['synset']} has particleApplier or particleSource with condition `toggled_on:True` but is not annotated as toggleable"
                                    formatted_conditions.append(("toggled_on", True))
                                else:
                                    raise ValueError(f"Synset {param_record['synset']} prop {prop} has unhandled condition {condition}")
                            formatted_param_value = {
                                param_record["system"]: formatted_conditions
                            }

                        else:
                            raise ValueError(f"prop {prop} not handled for parameter name `conditions`")
                
                    # Can skip system since it's just part of handling `conditions`
                    elif param_name == "system": continue
                    
                    # `method` values
                    elif param_name == "method": 
                        if (prop == "particleApplier") or (prop == "particleSource"):
                            formatted_param_value = "projection"
                        else:
                            raise ValueError(f"prop {prop} not handled for parameter name `method`")
                    
                    # Required derivative synsets
                    elif param_name == "meltable_derivative_synset":
                        formatted_param_value = param_value
                    
                    # Float values
                    else:
                        try: 
                            formatted_param_value = float(param_value)
                        except ValueError:
                            raise ValueError(f"Synset {param_record['synset']} property {prop} has param {param_name} that is not named `method`, `conditions`, or `system` and is not a NaN or a float. This is unhandled - please check.")

                    syns_to_props[param_record["synset"]][prop][param_name] = formatted_param_value
    

    with open(PARAMS_OUTFILE_FN, "w") as f:
        json.dump(syns_to_props, f, indent=4)
    
    print("Params parsed and added to flat and hierarchical files, saved.")
    return syns_to_props
    


if __name__ == "__main__":
    create_get_save_propagated_annots_params()




import csv 
import os

from tm_submap_params import TM_SUBMAPS_TO_PARAMS

for submap_fn in os.listdir("tm_raw_data_old"):
    with open("tm_raw_data_old/" + submap_fn, "r") as f:
        reader = csv.DictReader(f)
        rules = [rec for rec in reader]
    with open("tm_raw_data/" + submap_fn, "w") as f:
        fieldnames = ["rule_name"] + list(TM_SUBMAPS_TO_PARAMS[submap_fn.split(".csv")[0]].keys())
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writeheader()
        updated_rules = []
        for rule in rules: 
            updated_rule = {}
            for param, value in rule.items(): 
                # if param == "input_synsets":
                #     print(value)
                    # import sys; sys.exit()
                if param not in fieldnames: continue
                if param == "input_synsets":
                    input_objects = []
                    input_states = []
                    atoms = value.split(" and ")
                    # print(atoms); import sys; sys.exit()
                    for atom in atoms:
                        if "real " in atom:
                            input_objects.append(atom.split(" ")[-1])
                        else: 
                            input_states.append(atom)
                    updated_rule["input_synsets"] = " and ".join(input_objects)
                    updated_rule["input_states"] = " and ".join(input_states)
                elif param == "output_synsets":
                    output_objects = []
                    output_states = []
                    atoms = value.split(" and ")
                    for atom in atoms:
                        if "real " in atom:
                            output_objects.append(atom.split(" ")[-1])
                        else: 
                            output_states.append(atom)
                    updated_rule["output_synsets"] = " and ".join(output_objects)
                    updated_rule["output_states"] = " and ".join(output_states)
                elif param == "machine": 
                    pass
                else:
                    updated_rule[param] = value
            writer.writerow(updated_rule)




'''
Iteration one: 
Strict multi-way tree (without no count)
We consider each unique path to a synset "different", with the obvious
problem that each synset can have multiple hypernyms
'''
import collections
import json
import pandas as pd
import pathlib
from nltk.corpus import wordnet as wn

HIERARCHY_OUTPUT_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "output_hierarchy.json"
HIERARCHY_PROPERTIES_OUTPUT_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "output_hierarchy_properties.json"
CATEGORY_MAPPING_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "category_mapping.csv"
SUBSTANCE_MAPPING_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "substance_hyperparams.csv"
SYN_PROP_PARAM_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "propagated_annots_params.json"


def add_igibson_objects(node, synset_to_cat, synset_to_substance):
  '''
  Go through the hierarchy and add the words associated with the synsets as attributes.
  '''
  if node["name"] in synset_to_cat:
    node["categories"] = sorted(synset_to_cat[node["name"]])
  if node["name"] in synset_to_substance:
    node["substances"] = sorted(synset_to_substance[node["name"]])

  if "children" in node:
    for child_node in node["children"]:
      add_igibson_objects(child_node, synset_to_cat, synset_to_substance)


def add_path(path, hierarchy):
  if not path: 
    return 
  synset = path[-1]
  if "children" not in hierarchy:
    hierarchy["children"] = []
  for subhierarchy in hierarchy["children"]:
    if subhierarchy["name"] == synset:
      subhierarchy["hasModel"] = True 
      add_path(path[:-1], subhierarchy)
      return 
  hierarchy["children"].append({"name": synset, "hasModel": True})
  add_path(path[:-1], hierarchy["children"][-1])
  

def generate_paths(paths, path, synset, syn_prop_dict):
  # Annotated as not custom, or not present (i.e. not leaf, i.e. not custom)
  if (synset not in syn_prop_dict) or (not syn_prop_dict[synset]["is_custom"]):
    hypernyms = [x.name() for x in wn.synset(synset).hypernyms()]   # This fails iff a custom synset is incorrectly stored as non-custom
  else:
    hypernyms = syn_prop_dict[synset]["hypernyms"].split(",")

  if not hypernyms:
    paths.append(path)
  else:
    for hypernym in hypernyms:
      generate_paths(paths, path + [hypernym], hypernym, syn_prop_dict)


def add_properties(node, syn_prop_param_dict):
  node["abilities"] = syn_prop_param_dict[node["name"]]
  if "children" in node: 
    for child_node in node["children"]:
      add_properties(child_node, syn_prop_param_dict)


# API

def get_hierarchy(syn_prop_dict): 
  hierarchy = {"name": 'entity.n.01', "children": [], "hasModel": False}
  objects = set()
  for synset in syn_prop_dict.keys():
    # NOTE skip custom synsets without hypernym annotations
    if syn_prop_dict[synset]["is_custom"] and syn_prop_dict[synset]["hypernyms"] == "":
      continue
    objects.add(synset)
    paths = []
    generate_paths(paths, [synset], synset, syn_prop_dict)
    for path in paths: 
      add_path(path[:-1], hierarchy)

  synset_to_cat_raw = pd.read_csv(CATEGORY_MAPPING_FN)[["category", "synset"]].to_dict(orient="records")
  synset_to_cat = collections.defaultdict(list)
  for rec in synset_to_cat_raw: 
    synset_to_cat[rec["synset"]].append(rec["category"])

  synset_to_substance_raw = pd.read_csv(SUBSTANCE_MAPPING_FN)[["substance", "synset"]].to_dict(orient="records")
  synset_to_substance = collections.defaultdict(list)
  for rec in synset_to_substance_raw: 
    synset_to_substance[rec["synset"]].append(rec["substance"])

  add_igibson_objects(hierarchy, synset_to_cat, synset_to_substance)

  with open(HIERARCHY_OUTPUT_FN, "w") as f:
    json.dump(hierarchy, f, indent=2)
  return hierarchy


def create_get_save_hierarchy_with_properties(hierarchy):
  print("Adding params to hierarchy file...")
  with open(SYN_PROP_PARAM_FN) as f:
    syn_prop_param_dict = json.load(f)
  add_properties(hierarchy, syn_prop_param_dict)
  with open(HIERARCHY_PROPERTIES_OUTPUT_FN, "w") as f:
    json.dump(hierarchy, f, indent=2)
  print("Added params to hierarchy file, saved.")
  return hierarchy


if __name__ == "__main__":
  pass

"""
Graph using http://bl.ocks.org/eyaler/10586116#graph.json
"""


import os
import pathlib
import time
import gspread
import csv

OUTPUT_ROOT = pathlib.Path(__file__).parents[1] / "generated_data"

ASSETS_SHEET_KEY = "10L8wjNDvr1XYMMHas4IYYP9ZK7TfQHu--Kzoi0qhAe4"
SYNSETS_SHEET_KEY = "1eIQn1HzUJV15nCP4MqsHvrdWAV9VrKoxOqSnQxF0_1A"
SYNSET_PARAMS_SHEET_KEY = "1jQomcQS3DSMLctEOElafCPj_v498gwtb57NC9eNwu-4"
EXPLICIT_TRANSITION_RULES_SHEET_KEY = "1q3MqvnT_bOVbkMit1c7dhbzsOvd3KUQhJGx3aaJiH04"

ALL_SHEETS = [
  (ASSETS_SHEET_KEY, "Object Category Mapping", "category_mapping.csv"),
  (ASSETS_SHEET_KEY, "Allowed Room Types", "allowed_room_types.csv"),
  (ASSETS_SHEET_KEY, "Substance Hyperparams", "substance_hyperparams.csv"),
  (ASSETS_SHEET_KEY, "Object Renames", "object_renames.csv"),
  (ASSETS_SHEET_KEY, "Deletion Queue", "deletion_queue.csv"),
  (SYNSETS_SHEET_KEY, "Synsets", "synsets.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "heatSource", "prop_param_annots/heatSource.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "coldSource", "prop_param_annots/coldSource.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "cookable", "prop_param_annots/cookable.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "flammable", "prop_param_annots/flammable.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "heatable", "prop_param_annots/heatable.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "particleApplier", "prop_param_annots/particleApplier.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "particleSource", "prop_param_annots/particleSource.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "particleRemover", "prop_param_annots/particleRemover.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "particleSink", "prop_param_annots/particleSink.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "diceable", "prop_param_annots/diceable.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "sliceable", "prop_param_annots/sliceable.csv"),
  (SYNSET_PARAMS_SHEET_KEY, "meltable", "prop_param_annots/meltable.csv"),
  (EXPLICIT_TRANSITION_RULES_SHEET_KEY, "heat_cook", "transition_map/tm_raw_data/heat_cook.csv"),
  (EXPLICIT_TRANSITION_RULES_SHEET_KEY, "mixing_stick", "transition_map/tm_raw_data/mixing_stick.csv"),
  (EXPLICIT_TRANSITION_RULES_SHEET_KEY, "single_toggleable_machine", "transition_map/tm_raw_data/single_toggleable_machine.csv"),
  (EXPLICIT_TRANSITION_RULES_SHEET_KEY, "washer", "transition_map/tm_raw_data/washer.csv"),
]

def main():
  OUTPUT_ROOT.mkdir(exist_ok=True)
  client = gspread.service_account(filename=os.environ["GOOGLE_APPLICATION_CREDENTIALS"])
  for key, sheet_name, filename in ALL_SHEETS:
    worksheet = client.open_by_key(key).worksheet(sheet_name)
    with open(OUTPUT_ROOT / filename, "w") as f:
      writer = csv.writer(f)
      writer.writerows(worksheet.get_all_values())

    # Stop Google from complaining.
    time.sleep(10)

if __name__ == "__main__":
  main()


import csv
from collections import defaultdict, Counter
import logging
import json
import glob
import nltk
import pathlib
import bddl
from bddl.knowledge_base.orm import IntegrityError
from bddl.object_taxonomy import ObjectTaxonomy
from bddl.activity import Conditions, get_all_activities, get_instance_count
from bddl.config import get_definition_filename
import tqdm
from bddl.knowledge_base.models import *
from bddl.knowledge_base.utils import *


BDDL_DIR = pathlib.Path(bddl.__file__).parent
GENERATED_DATA_DIR = BDDL_DIR / "generated_data"

logger = logging.getLogger(__name__)

class KnowledgeBaseProcessor():
    def __init__(self, verbose=True) -> None:
        self._verbose = verbose

    def tqdm(self, iterable, *args, **kwargs):
        if self._verbose:
            return tqdm.tqdm(iterable, *args, **kwargs)
        else:
            return iterable
        
    def debug_print(self, *args, **kwargs):
        if self._verbose:
            print(*args, **kwargs)

    def run(self):
        logger.warning("Loading BDDL knowledge base... This may take a few seconds.")
        self.preparation()
        self.create_synsets()
        self.create_objects()
        self.create_scenes()
        self.create_tasks()
        self.create_transitions()
        self.post_complete_operation()
    

    # =============================== helper functions ===============================
    def preparation(self):
        """
        put any preparation work (e.g. sanity check) here
        """
        self.debug_print("Running preparation work...")

        nltk.download('wordnet')

        self.object_taxonomy = ObjectTaxonomy()

        # sanity check room types are up to date
        room_types_from_model = set([room_type for _, room_type in ROOM_TYPE_CHOICES])
        with open(GENERATED_DATA_DIR / 'allowed_room_types.csv', newline='') as csvfile:
            reader = csv.reader(csvfile, delimiter=',')
            room_types_from_csv = set([row[0] for row in reader][1:])
        assert room_types_from_model == room_types_from_csv, "room types are not up to date with allowed_room_types.csv"

        # get object rename mapping
        self.object_rename_mapping = {}
        self.obj_rename_mapping_unique_set = set()
        self.obj_rename_mapping_duplicate_set = set()
        with open(GENERATED_DATA_DIR / "object_renames.csv", newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                new_cat = row["New Category"].strip()
                obj_name = row["Object name"].strip()
                # sanity checks
                if obj_name != "":
                    assert len(obj_name.split('-')) == 2, f"{obj_name} should only have one \'-\'"
                    obj_id = obj_name.split('-')[1]
                    if obj_id not in self.obj_rename_mapping_unique_set:
                        self.obj_rename_mapping_unique_set.add(obj_id)
                    else:
                        self.obj_rename_mapping_duplicate_set.add(obj_id)
                    self.object_rename_mapping[obj_name] = f"{new_cat}-{obj_id}"
            assert len(self.obj_rename_mapping_duplicate_set) == 0, f"object rename mapping have duplicates: {self.obj_rename_mapping_duplicate_set}"

        self.debug_print("Finished prep work...")

    def post_complete_operation(self):
        """
        put any post completion work (e.g. update stuff) here
        """
        self.debug_print("Running post completion operations...")
        self.generate_synset_state()
        # self.generate_object_images()
        # self.nuke_unused_synsets()


    def create_synsets(self):
        """
        create synsets with annotations from propagated_annots_canonical.json and hierarchy from output_hierarchy.json
        """
        from nltk.corpus import wordnet as wn

        self.debug_print("Creating synsets...")
        for synset_name in self.tqdm(self.object_taxonomy.get_all_synsets()):
            synset_is_custom = not wn_synset_exists(synset_name)  # TODO: use data from hierarchy. synset_sub_hierarchy["is_custom"] == "1"
            if synset_name != canonicalize(synset_name):
                self.debug_print(f"synset {synset_name} is not canonicalized!")
            synset_definition = wn.synset(synset_name).definition() if wn_synset_exists(synset_name) else ""
            synset, created = Synset.get_or_create(name=synset_name, defaults={"definition": synset_definition, "is_custom": synset_is_custom})
            parents = self.object_taxonomy.get_parents(synset_name)
            for parent in parents:
                parent_obj = Synset.get(name=parent)
                synset.parents.add(parent_obj)
            cur_ancestors = self.object_taxonomy.get_ancestors(synset_name)
            for ancestor in sorted(cur_ancestors):
                ancestor_obj = Synset.get(name=ancestor)
                synset.ancestors.add(ancestor_obj)

            # Add any categories
            for category in self.object_taxonomy.get_categories(synset_name):
                assert not any(c.name == category for c in Category.all_objects()), f"Category {category} of {synset_name} already exists!"
                Category.get_or_create(name=category, synset=synset)

            # Add any properties
            if created:
                for property_name, params in self.object_taxonomy.get_abilities(synset_name).items():
                    Property.create(synset=synset, name=property_name, parameters=json.dumps(params))
        

    def create_objects(self):
        """
        Create objects and map to categories (with object inventory)
        """
        self.debug_print("Creating objects...")
        # first get Deletion Queue
        self.deletion_queue = set()
        with open(GENERATED_DATA_DIR / "deletion_queue.csv", newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                self.deletion_queue.add(row["Object"].strip().split("-")[1])
        # then create objects
        with open(GENERATED_DATA_DIR / "object_inventory_future.json", "r") as f:
            inventory = json.load(f)
            for orig_name, provider in self.tqdm(inventory["providers"].items()):
                object_name = self.object_rename_mapping[orig_name] if orig_name in self.object_rename_mapping else orig_name
                if object_name.split("-")[1] in self.deletion_queue:
                    continue

                # Create the object
                category_name = object_name.split("-")[0]
                category, _ = Category.get_or_create(name=category_name)
                object = Object.create(name=object_name, original_name=orig_name, ready=False, provider=provider, category=category)
                if orig_name in inventory["meta_links"]:
                    for meta_link in inventory["meta_links"][orig_name]:
                        meta_link_obj, _ = MetaLink.get_or_create(name=meta_link)
                        object.meta_links.add(meta_link_obj)

        with open(GENERATED_DATA_DIR / "object_inventory.json", "r") as f:
            objs = []
            for orig_name in self.tqdm(json.load(f)["providers"].keys()):
                object_name = self.object_rename_mapping[orig_name] if orig_name in self.object_rename_mapping else orig_name
                if object_name.split("-")[1] not in self.deletion_queue:
                    category_name = object_name.split("-")[0]
                    category, _ = Category.get_or_create(name=category_name)
                    # safeguard to ensure currently available objects are also in future planned dataset
                    assert Object.exists(name=object_name), f"{object_name} in category {category}, which exists in object_inventory.json, is not in object_inventory_future.json!"
                    object = Object.get(name=object_name)
                    object.ready = True
                    objs.append(object)


    def create_scenes(self):
        """
        create scene objects (which stores the room config)
        scene matching to tasks will be generated later when creating task objects
        """
        self.debug_print("Creating scenes...")
        with open(GENERATED_DATA_DIR / "combined_room_object_list_future.json", "r") as f:
            planned_scene_dict = json.load(f)["scenes"]
            for scene_name in self.tqdm(planned_scene_dict):
                scene, _ = Scene.get_or_create(name=scene_name)
                for room_name in planned_scene_dict[scene_name]:
                    # TODO: Implement unique_together constraints
                    try:
                        room = Room.create(
                            name=room_name, 
                            type=room_name.rsplit('_', 1)[0], 
                            ready=False, 
                            scene=scene
                        )
                    except IntegrityError:
                        raise Exception(f"room {room_name} in {scene.name} (not ready) already exists!")
                    for orig_name, count in planned_scene_dict[scene_name][room_name].items():
                        if orig_name.split("-")[1] not in self.deletion_queue:
                            object_name = self.object_rename_mapping[orig_name] if orig_name in self.object_rename_mapping else orig_name
                            object, _ = Object.get_or_create(name=object_name, defaults={
                                "original_name": orig_name,
                                "ready": False,
                                "planned": False,
                                "category": Category.get(name=object_name.split("-")[0]),
                            })
                            RoomObject.create(room=room, object=object, count=count)

        with open(GENERATED_DATA_DIR / "combined_room_object_list.json", "r") as f:
            current_scene_dict = json.load(f)["scenes"]
            for scene_name in self.tqdm(current_scene_dict):
                scene, _ = Scene.get_or_create(name=scene_name)
                for room_name in current_scene_dict[scene_name]:
                    try:
                        room = Room.create(
                            name=room_name, 
                            type=room_name.rsplit('_', 1)[0], 
                            ready=True, 
                            scene=scene
                        )
                    except IntegrityError:
                        raise Exception(f"room {room_name} in {scene.name} (ready) already exists!")
                    for orig_name, count in current_scene_dict[scene_name][room_name].items():
                        if orig_name.split("-")[1] not in self.deletion_queue:
                            object_name = self.object_rename_mapping[orig_name] if orig_name in self.object_rename_mapping else orig_name
                            object, _ = Object.get_or_create(name=object_name, defaults={
                                "original_name": orig_name,
                                "ready": False,
                                "planned": False,
                                "category": Category.get(name=object_name.split("-")[0])
                            })
                            RoomObject.create(room=room, object=object, count=count)


    def create_tasks(self):
        """
        create tasks and map to synsets
        """
        self.debug_print("Creating tasks...")
        tasks = glob.glob(str(BDDL_DIR / "activity_definitions/*"))
        tasks = [(act, inst) for act in get_all_activities() for inst in range(get_instance_count(act))]
        for act, inst in self.tqdm(tasks):
            # Load task definition
            conds = Conditions(act, inst, "omnigibson")
            synsets = set(synset for synset in conds.parsed_objects if synset != "agent.n.01")
            canonicalized_synsets = set(canonicalize(synset) for synset in synsets)
            with open(get_definition_filename(act, inst), "r") as f:
                raw_task_definition = "".join(f.readlines())

            initial_conds, goal_conds = get_initial_and_goal_conditions(conds)
            combined_conds = initial_conds + goal_conds

            # Create task object
            task_name = f"{act}-{inst}"
            task = Task.create(name=task_name, definition=raw_task_definition)
            for predicate in all_task_predicates(combined_conds):
                pred_obj, _ = Predicate.get_or_create(name=predicate)
                task.uses_predicates.add(pred_obj)

            # add any synset that is not currently in the database
            for synset_name in sorted(canonicalized_synsets):
                is_used_as_non_substance, is_used_as_substance = object_substance_match(combined_conds, synset_name)
                is_used_as_fillable = object_used_as_fillable(combined_conds, synset_name)
                # all annotated synsets have been created before, so any newly created synset is illegal
                synset, _ = Synset.get_or_create(name=synset_name)
                synset.is_used_as_substance = synset.is_used_as_substance or is_used_as_substance
                synset.is_used_as_non_substance = synset.is_used_as_non_substance or is_used_as_non_substance
                synset.is_used_as_fillable = synset.is_used_as_fillable or is_used_as_fillable
                synset_used_predicates = object_used_predicates(combined_conds, synset_name)
                if not synset_used_predicates:
                    self.debug_print(f"Synset {synset_name} is not used in any predicate in {task_name}")
                for predicate in synset_used_predicates:
                    pred_obj, _ = Predicate.get_or_create(name=predicate)
                    if pred_obj not in synset.used_in_predicates:
                        synset.used_in_predicates.add(pred_obj)
                task.synsets.add(synset)

                # If the synset ever shows up as future or real, check validity
                used_as_future_or_real = "future" in synset_used_predicates or "real" in synset_used_predicates
                if used_as_future_or_real:
                    # Assert that it's used as future in initial and as real in goal
                    initial_preds = object_used_predicates(initial_conds, synset_name)
                    if "future" not in initial_preds:
                        self.debug_print(f"Synset {synset_name} is not used as future in initial in {task_name}")
                    if "real" in initial_preds:
                        self.debug_print(f"Synset {synset_name} is used as real in initial in {task_name}")

                    goal_preds = object_used_predicates(goal_conds, synset_name)
                    if "real" not in goal_preds:
                        self.debug_print(f"Synset {synset_name} is not used as real in goal in {task_name}")
                    if "future" in goal_preds:
                        self.debug_print(f"Synset {synset_name} is used as future in goal in {task_name}")

                    task.future_synsets.add(synset)

            # generate room requirements for task
            room_synset_requirements = defaultdict(Counter)  # room[synset] = count
            for cond in leaf_inroom_conds(conds.parsed_initial_conditions, synsets):
                assert len(cond) == 2, f"{task_name}: {str(cond)} not in correct format"
                rr_type = cond[1]
                rr_synset = cond[0]
                room_synset_requirements[rr_type][rr_synset] += 1

            for rr_type, synset_counter in room_synset_requirements.items():
                room_requirement = RoomRequirement.create(task=task, type=rr_type)
                for rsr_synset, count in synset_counter.items():
                    rsr_synset_obj = Synset.get(name=rsr_synset)
                    RoomSynsetRequirement.create(room_requirement=room_requirement, synset=rsr_synset_obj, count=count)


    def create_transitions(self):
        # Load the transition data
        json_paths = glob.glob(str(GENERATED_DATA_DIR / "transition_map/tm_jsons/*.json"))
        transitions = []
        for jp in json_paths:
            if "washer_" in jp:
                continue
            with open(jp) as f:
                transitions.extend(json.load(f))

        # Create the transition objects
        for transition_data in self.tqdm(transitions):
            rule_name = transition_data["rule_name"]
            transition = TransitionRule.create(name=rule_name)
            inputs = set(transition_data["input_synsets"].keys())
            assert inputs, f"Transition {transition.name} has no inputs!"
            outputs = set(transition_data["output_synsets"].keys())
            assert outputs, f"Transition {transition.name} has no outputs!"
            assert inputs & outputs == set(), f"Inputs and outputs of {transition.name} overlap!"
            for synset_name in inputs:
                synset = Synset.get(name=synset_name)
                transition.input_synsets.add(synset)
            for synset_name in outputs:
                synset = Synset.get(name=synset_name)
                transition.output_synsets.add(synset)

    def generate_synset_state(self):
        synsets = []
        substances = {s.name for s in Synset.all_objects() if "substance" in s.property_names}
        for synset in self.tqdm(Synset.all_objects()):
            if synset.name == "entity.n.01": synset.state = STATE_MATCHED   # root synset is always legal
            elif synset.name in substances:
                synset.state = STATE_SUBSTANCE
            elif synset.parents:
                if len(synset.matching_ready_objects) > 0:
                    synset.state = STATE_MATCHED
                elif len(synset.matching_objects) > 0:
                    synset.state = STATE_PLANNED
                else:
                    synset.state = STATE_UNMATCHED
            else:
                synset.state = STATE_ILLEGAL
            synsets.append(synset)


    def nuke_unused_synsets(self):
        # Make repeated passes until we propagate far enough up
        while True:
            removal_names = set()
            for synset in Synset.all_objects():
                # In a given pass, only leaf nodes can be removed
                if synset.children:
                    continue

                # If a synset has objects or task relevance, we can't remove it
                if len(synset.matching_objects) != 0:
                    continue
                
                if synset.n_task_required != 0:
                    continue

                if synset.used_by_transition_rules:
                    continue

                if synset.produced_by_transition_rules:
                    continue

                # Otherwise queue it for removal
                removal_names.add(synset.name)

            if removal_names:
                for s in list(Synset.all_objects()):
                    if s.name in removal_names:
                        s.delete()
            else:
                break



if __name__ == "__main__":
    import IPython
    KnowledgeBaseProcessor().run()
    IPython.embed()


import pathlib
import json 
from nltk.corpus import wordnet as wn

PROPAGATED_CANONICAL = pathlib.Path(__file__).parents[1] / "generated_data" / "propagated_annots_canonical.json"


def get_intersection(synset, property_map, canonical):
    if "children" in synset: # if synset node has children
        props = set()
        for i in range(len(synset["children"])):
            child_synset = synset["children"][i]
            child_props = get_intersection(child_synset, property_map, canonical)
            if i == 0:
                props = set(child_props)
            else:
                props = props.intersection(child_props)
        property_map[synset["name"]] = props # add these once they have been intersected
        return props
    else: # base case
        if synset["name"] in canonical: # if the synsetified name shows up in canonical
            props = canonical[synset["name"]] # get its properties
            property_map[synset["name"]] = props
        else: # if the synsetified name isn't in canonical
            for elem in canonical: # check every synset of canonical to see if it's a synonym of variable `synset`
                try: # handle crashes from trying to iterate through custom synsets
                    x = wn.synset(elem)
                except:
                    pass
                else:
                    if wn.synset(elem) == wn.synset(synset["name"]): # if there's a synonym
                        props = canonical[elem] # append its properties
                        property_map[synset["name"]] = props
        return props
      

def process_output(prop_map):
    dict = {}
    for synset in sorted(prop_map):
        dict[synset] = {}
        for elem in sorted(prop_map[synset]):
            dict[synset][elem] = {}
    return dict


# API

def create_get_save_propagated_canonical(hierarchy, annots_canonical):
    prop_map = {}
    get_intersection(hierarchy, prop_map, annots_canonical)
    made_dict = process_output(prop_map)
    with open(PROPAGATED_CANONICAL, "w") as f:
      json.dump(made_dict, f, indent=2)
    return made_dict
    

if __name__ == '__main__':
    pass



from bddl.knowledge_base.models import Property, \
  MetaLink, \
  Predicate, \
  Scene, \
  Category, \
  Object, \
  Synset, \
  TransitionRule, \
  Task, \
  RoomRequirement, \
  RoomSynsetRequirement, \
  Room, \
  RoomObject

from bddl.knowledge_base.utils import STATE_MATCHED, \
  STATE_PLANNED, \
  STATE_UNMATCHED, \
  STATE_SUBSTANCE, \
  STATE_ILLEGAL, \
  STATE_NONE

from bddl.knowledge_base.processing import KnowledgeBaseProcessor

# Load the knowledge base
KnowledgeBaseProcessor(verbose=False).run()


import os

# PATHS
ACTIVITY_CONFIGS_PATH = os.path.join(os.path.dirname(__file__), 'activity_definitions')

# BDDL
SUPPORTED_BDDL_REQUIREMENTS = [':strips', ':negative-preconditions', ':typing', ':adl']

READABLE_PREDICATE_NAMES = {
    'ontop': 'on top of',
    'nextto': 'next to'
}


def get_definition_filename(behavior_activity, instance, domain=False):
    if domain:
        return os.path.join(ACTIVITY_CONFIGS_PATH, 'domain_igibson.bddl')
    else:
        return os.path.join(ACTIVITY_CONFIGS_PATH, behavior_activity, f"problem{instance}.bddl")


def get_domain_filename(domain_name):
    return os.path.join(ACTIVITY_CONFIGS_PATH, f"domain_{domain_name}.bddl")


# MISC
GROUND_GOALS_MAX_OPTIONS = 20
GROUND_GOALS_MAX_PERMUTATIONS = 10



from functools import cached_property, cache
from collections import defaultdict
from dataclasses import field, fields
import uuid

class IntegrityError(Exception):
    pass

class Model:
    _CLASS_REGISTRY = dict()
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Model._CLASS_REGISTRY[cls.__name__] = cls

    @staticmethod
    def get_model_class(target_model):
        return Model._CLASS_REGISTRY[target_model]

    _OBJECT_REGISTRY = defaultdict(dict)
    def __post_init__(self):
        key = self._key
        assert key not in self._OBJECT_REGISTRY[self.__class__.__name__], f"Duplicate key {key} for {self.__class__.__name__}"

        self._OBJECT_REGISTRY[self.__class__.__name__][key] = self

        # Also bind the "this_instance" member of each fk field
        for field in fields(self):
            if field.type in (ManyToOne, OneToMany, ManyToMany):
                fk = getattr(self, field.name)
                fk._this_inst = self

    @cached_property
    def _key(self):
        return getattr(self, self.Meta.pk)

    @classmethod
    def get(cls, *args, **kwargs):
        if args:
            assert not kwargs, "Cannot specify both positional and keyword arguments to get"
            pk, = args
            kwargs = {cls.Meta.pk: pk}
        assert kwargs, "Must specify either positional or keyword arguments to get"
        # Use a fast lookup for the PK.
        if len(kwargs) == 1 and cls.Meta.pk in kwargs:
            key = kwargs[cls.Meta.pk]
            if key is None:
                return None
            return cls._OBJECT_REGISTRY[cls.__name__][key] if key in cls._OBJECT_REGISTRY[cls.__name__] else None
        objs = [x for x in cls.all_objects() if all(getattr(x, attr) == value for attr, value in kwargs.items())]
        if not objs:
            return None
        assert len(objs) == 1, f"Expected exactly one object matching {kwargs} but found {objs}"
        return objs[0]
    
    @classmethod
    def exists(cls, *args, **kwargs):
        if args:
            assert not kwargs, "Cannot specify both positional and keyword arguments to get"
            pk, = args
            kwargs = {cls.Meta.pk: pk}
        assert kwargs, "Must specify either positional or keyword arguments to get"
        # Use a fast lookup for the PK.
        if len(kwargs) == 1 and cls.Meta.pk in kwargs:
            return kwargs[cls.Meta.pk] in cls._OBJECT_REGISTRY[cls.__name__]
        objs = [x for x in cls.all_objects() if all(getattr(x, attr) == value for attr, value in kwargs.items())]
        return bool(objs)
    
    @classmethod
    def get_or_create(cls, defaults=None, *args, **kwargs):
        if cls.exists(*args, **kwargs):
            return cls.get(*args, **kwargs), False
        else:
            create_args = {}
            if defaults:
                create_args.update(defaults)
            if args:
                assert not kwargs, "Cannot specify both positional and keyword arguments to get"
                pk, = args
                create_args[cls.Meta.pk] = pk
            if kwargs:
                create_args.update(kwargs)
            return cls.create(**create_args), True

    @classmethod
    def create(cls, **kwargs):
        # Check the kwargs to see if any are foreign keys
        sanitized_kwargs = {k: v for k, v in kwargs.items() if k not in cls.foreign_key_field_names()}
        obj = cls(**sanitized_kwargs)

        foreign_kwargs = {k + "_fk": v for k, v in kwargs.items() if k in cls.foreign_key_field_names()}
        for field_name, field_value in foreign_kwargs.items():
            fk = getattr(obj, field_name)
            if isinstance(fk, ManyToOne):
                fk.set(field_value)
            elif isinstance(fk, ManyToMany):
                for obj in field_value:
                    fk.add(obj)
            elif isinstance(fk, OneToMany):
                for obj in field_value:
                    fk.add(obj)
                fk.add(obj, field_value)
            else:
                raise ValueError(f"Unknown foreign key type {fk}")
            
        # Build the unique-together key
        # TODO: Make this work beyond just creation.
        # if hasattr(cls.Meta, "unique_together"):
        #     unique_together_keys = cls.Meta.unique_together
        #     this_key = tuple([getattr(obj, attr) for attr in unique_together_keys])
        #     for other in cls._OBJECT_REGISTRY[cls.__name__].values():
        #         if other is obj:
        #             continue
        #         other_key = tuple([getattr(other, attr) for attr in unique_together_keys])
        #         if this_key == other_key:
        #             raise IntegrityError(f"Unique together constraint violated for {cls.__name__}. New item: {obj!r}. Existing item: {other!r}. Uniqueness constraint: {unique_together_keys}")

        return obj

    @classmethod
    def foreign_key_field_names(cls):
        fk_fields = [field.name for field in fields(cls) if field.type in (ManyToOne, OneToMany, ManyToMany)]
        assert all(field.endswith("_fk") for field in fk_fields), f"Foreign key fields must end with _fk. Fields: {fk_fields}"
        return [field[:-3] for field in fk_fields]

    @classmethod
    def all_objects(cls):
        return sorted(x for x in cls._OBJECT_REGISTRY[cls.__name__].values())
    
    def __lt__(self, other):
        if hasattr(self.Meta, "ordering"):
            order_attrs = self.Meta.ordering
        else:
            order_attrs = [self.Meta.pk]
        self_order_key = [getattr(self, attr) for attr in order_attrs]
        other_order_key = [getattr(other, attr) for attr in order_attrs]
        return self_order_key < other_order_key

    def __getattr__(self, key):
        try:
            fk = self.__getattribute__(key + "_fk")
            if hasattr(fk, "get"):  # ManyToOne
                return fk.get()
            else:
                return fk
        except AttributeError:
            return self.__getattribute__(key)

    def __setattr__(self, key, value):
        try:
            fk = self.__getattribute__(key + "_fk")
            if hasattr(fk, "set"):  # ManyToOne
                return fk.set(self, value)
            else:
                raise ValueError(f"Cannot set {key} because it is a OneToMany or ManyToMany field.")
        except AttributeError:
            return super().__setattr__(key, value)

    def __hash__(self) -> int:
        return hash(self._key)
    
    def __eq__(self, __value: object) -> bool:
        return __value is self

def has_field_with_type(cls, field_name, field_type):
    cls_fields = [x for x in fields(cls) if x.name == field_name]
    if not cls_fields:
        return False
    field, = cls_fields
    return field.type == field_type

def get_model_class(target_model):
    if isinstance(target_model, str):
        return Model.get_model_class(target_model)
    elif isinstance(target_model, type) and issubclass(target_model, Model):
        return target_model
    else:
        raise ValueError(f"Invalid target model {target_model}")

class ManyToOne:
    def __init__(self, target_model, target_field):
        target_model = get_model_class(target_model)
        target_field = target_field + "_fk"
        assert has_field_with_type(target_model, target_field, OneToMany), f"Target model {target_model} does not have a OneToMany field {target_field}. Fields: {[x for x in fields(target_model)]}"

        self._target_model = target_model
        self._target_field = target_field

        self._target_key = None
        self._this_inst = None
    
    def get(self):
        assert self._this_inst is not None, "Unbound foreign key object"
        return self._target_model.get(self._target_key)
    
    def set(self, target_inst):
        assert self._target_key is None, "Foreign key object already bound"
        assert self._this_inst is not None, "Unbound foreign key object"
        assert target_inst.__class__ == self._target_model, f"Target instance {target_inst} is not of type {self._target_model}"
        self._target_key = target_inst._key
        assert self._this_inst._key not in getattr(target_inst, self._target_field)._values, f"Foreign key constraint violated: {self._this_inst} already has a foreign key to {target_inst}"
        getattr(target_inst, self._target_field)._values.append(self._this_inst._key)

class OneToMany:
    def __init__(self, target_model, target_field) -> None:
        target_model = get_model_class(target_model)
        target_field = target_field + "_fk"
        assert has_field_with_type(target_model, target_field, ManyToOne), f"Target model {target_model} does not have a ManyToOne field {target_field}. Fields: {[x for x in fields(target_model)]}"

        self._target_model = target_model
        self._target_field = target_field

        self._values = []
        self._this_inst = None

    def __iter__(self):
        assert self._this_inst is not None, "Unbound foreign key object"
        return (self._target_model.get(key) for key in self._values)
    
    def __len__(self):
        assert self._this_inst is not None, "Unbound foreign key object"
        return len(self._values)
    
    def __getitem__(self, key):
        assert self._this_inst is not None, "Unbound foreign key object"
        if isinstance(key, slice):
            return (self._target_model.get(key) for key in self._values[key])
        else:
            return self._target_model.get(self._values[key])
    
    def add(self, target_inst):
        # Directly call their setter for verification
        assert self._this_inst is not None, "Unbound foreign key object"
        getattr(target_inst, self._target_field).set(self._this_inst._key)

class ManyToMany:
    def __init__(self, target_model, target_field) -> None:
        target_model = get_model_class(target_model)
        target_field = target_field + "_fk"
        assert has_field_with_type(target_model, target_field, ManyToMany), f"Target model {target_model} does not have a ManyToMany field {target_field}. Fields: {[x for x in fields(target_model)]}"

        self._target_model = target_model
        self._target_field = target_field

        self._values = []
        self._this_inst = None

    def __iter__(self):
        assert self._this_inst is not None, "Unbound foreign key object"
        return (self._target_model.get(key) for key in self._values)
    
    def __getitem__(self, key):
        assert self._this_inst is not None, "Unbound foreign key object"
        if isinstance(key, slice):
            return (self._target_model.get(key) for key in self._values[key])
        else:
            return self._target_model.get(self._values[key])

    def __len__(self):
        assert self._this_inst is not None, "Unbound foreign key object"
        return len(self._values)

    def add(self, target_inst):
        assert self._this_inst is not None, "Unbound foreign key object"
        assert target_inst.__class__ == self._target_model, f"Target instance {target_inst} is not of type {self._target_model}"
        assert target_inst._key not in self._values, f"Foreign key constraint violated: {self._this_inst} already has a foreign key to {target_inst}"
        assert self._this_inst._key not in getattr(target_inst, self._target_field)._values, f"Foreign key constraint violated: {self._this_inst} already has a foreign key to {target_inst}"
        self._values.append(target_inst._key)
        getattr(target_inst, self._target_field)._values.append(self._this_inst._key)

def ManyToOneField(target_model, target_field, **kwargs):
    return field(default_factory=lambda: ManyToOne(target_model, target_field), repr=False, compare=False, **kwargs)

def OneToManyField(target_model, target_field, **kwargs):
    return field(default_factory=lambda: OneToMany(target_model, target_field), repr=False, compare=False, **kwargs)

def ManyToManyField(target_model, target_field, **kwargs):
    return field(default_factory=lambda: ManyToMany(target_model, target_field), repr=False, compare=False, **kwargs)

def UUIDField(**kwargs):
    return field(default_factory=lambda: str(uuid.uuid4()), **kwargs)


from setuptools import setup, find_packages

# TODO package data

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='bddl',
    version='3.4.0b4',
    author='Stanford University',
    long_description_content_type='text/markdown',
    long_description=long_description,
    url='https://github.com/StanfordVL/bddl',
    zip_safe=False,
    packages=find_packages(),
    install_requires=[
        'pytest',
        'numpy',
        'networkx',
        'jupytext',
        'future',
        'nltk~=3.7'
    ],
    package_data={},
    include_package_data=True
)



from bddl.logic_base import UnaryAtomicFormula, BinaryAtomicFormula
from bddl.backend_abc import BDDLBackend
from bddl.parsing import parse_domain


*__, domain_predicates = parse_domain("omnigibson")
UNARIES = [predicate for predicate, inputs in domain_predicates.items() if len(inputs) == 1]
BINARIES = [predicate for predicate, inputs in domain_predicates.items() if len(inputs) == 2]


class TrivialBackend(BDDLBackend):
    def get_predicate_class(self, predicate_name):
        PREDICATE_MAPPING = {
            "cooked": TrivialCookedPredicate,
            "frozen": TrivialFrozenPredicate,
            "open": TrivialOpenPredicate,
            "folded": TrivialFoldedPredicate,
            "unfolded": TrivialUnfoldedPredicate,
            "toggled_on": TrivialToggledOnPredicate,
            "hot": TrivialHotPredicate,
            "frozen": TrivialFrozenPredicate,
            "on_fire": TrivialOnFirePredicate,
            "empty": TrivialEmptyPredicate,
            "closed": TrivialClosedPredicate,
            "future": TrivialFuturePredicate,
            "real": TrivialRealPredicate,
            "covered": TrivialCoveredPredicate,
            "ontop": TrivialOnTopPredicate,
            "inside": TrivialInsidePredicate,
            "filled": TrivialFilledPredicate,
            "saturated": TrivialSaturatedPredicate,
            "contains": TrivialContainsPredicate,
            "ontop": TrivialOnTopPredicate,
            "nextto": TrivialNextToPredicate,
            "under": TrivialUnderPredicate,
            "touching": TrivialTouchingPredicate,
            "overlaid": TrivialOverlaidPredicate,
            "attached": TrivialAttachedPredicate,
            "draped": TrivialDrapedPredicate,
            "insource": TrivialInsourcePredicate,
            "broken": TrivialBrokenPredicate,
            "assembled": TrivialAssembledPredicate,
            "grasped": TrivialGraspedPredicate,
        } 
        return PREDICATE_MAPPING[predicate_name]


class TrivialSimulator(object):
    def __init__(self):
        # Unaries - populated with 1-tuples of string names 
        self.cooked = set()
        self.frozen = set()
        self.open = set()
        self.folded = set()
        self.unfolded = set()
        self.toggled_on = set() 
        self.hot = set() 
        self.on_fire = set() 
        self.empty = set() 
        self.future = set() 
        self.real = set() 
        self.broken = set()
        self.closed = set()
        self.assembled = set()
        # Binaries - populated with 2-tuples of string names
        self.saturated = set()
        self.covered = set() 
        self.filled = set() 
        self.contains = set() 
        self.ontop = set() 
        self.nextto = set() 
        self.under = set() 
        self.touching = set() 
        self.inside = set() 
        self.overlaid = set() 
        self.attached = set() 
        self.draped = set() 
        self.insource = set() 
        self.grasped = set()

        self.create_predicate_to_setters()
    
    def create_predicate_to_setters(self):
        self.predicate_to_setters = {
            "cooked": self.set_cooked,
            "frozen": self.set_frozen,
            "open": self.set_open,
            "folded": self.set_folded,
            "unfolded": self.set_unfolded,
            "toggled_on": self.set_toggled_on,
            "hot": self.set_hot,
            "on_fire": self.set_on_fire,
            "empty": self.set_empty,
            "broken": self.set_broken,
            "assembled": self.set_assembled,
            "closed": self.set_closed,
            "future": self.set_future,
            "real": self.set_real,
            "inside": self.set_inside,
            "ontop": self.set_ontop,
            "covered": self.set_covered,
            "filled": self.set_filled,
            "saturated": self.set_saturated,
            "nextto": self.set_nextto,
            "contains": self.set_contains,
            "under": self.set_under,
            "touching": self.set_touching,
            "overlaid": self.set_overlaid,
            "attached": self.set_attached,
            "draped": self.set_draped,
            "insource": self.set_insource,
            "grasped": self.set_grasped,
        }

    def set_state(self, literals): 
        """
        Given a set of non-contradictory parsed ground literals, set this backend to them. 
        Also set implied predicates:
            filled => contains
            not contains => not filled 
            ontop => nextto
            not nextto => not ontop
            TODO under? others? draped?
        """
        for literal in literals: 
            is_predicate = not(literal[0] == "not")
            predicate, *objects = literal[1] if (literal[0] == "not") else literal
            if predicate == "inroom": 
                print(f"Skipping inroom literal {literal}")
                continue
            self.predicate_to_setters[predicate](tuple(objects), is_predicate)
            # Entailed predicates 
            if is_predicate and (predicate == "filled"):
                self.predicate_to_setters["contains"](tuple(objects), True)
            if (not is_predicate) and (predicate == "contains"):
                self.predicate_to_setters["filled"](tuple(objects), False)
            if is_predicate and (predicate == "ontop"):
                self.predicate_to_setters["nextto"](tuple(objects), True)
            if (not is_predicate) and (predicate == "nextto"):
                self.predicate_to_setters["ontop"](tuple(objects), False)

    def set_cooked(self, objs, is_cooked):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_cooked: 
            self.cooked.add(objs)
        else: 
            self.cooked.discard(objs)
    
    def get_cooked(self, objs):
        return tuple(obj.name for obj in objs) in self.cooked
    
    def set_frozen(self, objs, is_frozen):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_frozen: 
            self.frozen.add(objs)
        else: 
            self.frozen.discard(objs)
    
    def get_frozen(self, objs):
        return tuple(obj.name for obj in objs) in self.frozen
    
    def set_open(self, objs, is_open):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_open: 
            self.open.add(objs)
        else: 
            self.open.discard(objs)
    
    def get_open(self, objs):
        return tuple(obj.name for obj in objs) in self.open
    
    def set_folded(self, objs, is_folded):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_folded: 
            self.folded.add(objs)
        else: 
            self.folded.discard(objs)
    
    def get_folded(self, objs):
        return tuple(obj.name for obj in objs) in self.folded
    
    def set_unfolded(self, objs, is_unfolded):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_unfolded: 
            self.unfolded.add(objs)
        else: 
            self.unfolded.discard(objs)
    
    def get_unfolded(self, objs):
        return tuple(obj.name for obj in objs) in self.unfolded
    
    def set_toggled_on(self, objs, is_toggled_on):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_toggled_on: 
            self.toggled_on.add(objs)
        else: 
            self.toggled_on.discard(objs)
    
    def get_toggled_on(self, objs):
        return tuple(obj.name for obj in objs) in self.toggled_on
    
    def set_hot(self, objs, is_hot):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_hot: 
            self.hot.add(objs)
        else: 
            self.hot.discard(objs)
    
    def get_hot(self, objs):
        return tuple(obj.name for obj in objs) in self.hot
    
    def set_on_fire(self, objs, is_on_fire):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_on_fire: 
            self.on_fire.add(objs)
        else: 
            self.on_fire.discard(objs)
    
    def get_on_fire(self, objs):
        return tuple(obj.name for obj in objs) in self.on_fire
    
    def set_empty(self, objs, is_empty):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_empty: 
            self.empty.add(objs)
        else: 
            self.empty.discard(objs)
    
    def get_empty(self, objs):
        return tuple(obj.name for obj in objs) in self.empty
    
    def set_closed(self, objs, is_closed):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_closed: 
            self.closed.add(objs)
        else: 
            self.closed.discard(objs)
    
    def get_closed(self, objs):
        return tuple(obj.name for obj in objs) in self.closed
    
    def set_broken(self, objs, is_broken):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_broken: 
            self.broken.add(objs)
        else: 
            self.broken.discard(objs)
    
    def get_broken(self, objs):
        return tuple(obj.name for obj in objs) in self.broken
    
    def set_assembled(self, objs, is_assembled):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_assembled: 
            self.assembled.add(objs)
        else: 
            self.assembled.discard(objs)
    
    def get_assembled(self, objs):
        return tuple(obj.name for obj in objs) in self.assembled
    
    def set_future(self, objs, is_future):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_future: 
            self.future.add(objs)
        else: 
            self.future.discard(objs)
    
    def get_future(self, objs):
        return tuple(obj.name for obj in objs) in self.future
    
    def set_real(self, objs, is_real):
        assert len(objs) == 1, f"`objs` has len other than 1: {objs}"
        if is_real: 
            self.real.add(objs)
        else: 
            self.real.discard(objs)
    
    def get_real(self, objs):
        return tuple(obj.name for obj in objs) in self.real

    def set_covered(self, objs, is_covered):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_covered:
            self.covered.add(objs)
        else:
            self.covered.discard(objs)

    def get_covered(self, objs):
        return tuple(obj.name for obj in objs) in self.covered
    
    def set_ontop(self, objs, is_ontop):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_ontop:
            self.ontop.add(objs)
        else:
            self.ontop.discard(objs)

    def get_ontop(self, objs):
        return tuple(obj.name for obj in objs) in self.ontop
    
    def set_inside(self, objs, is_inside):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_inside:
            self.inside.add(objs)
        else:
            self.inside.discard(objs)
    
    def get_inside(self, objs):
        return tuple(obj.name for obj in objs) in self.inside
    
    def set_filled(self, objs, is_filled):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_filled:
            self.filled.add(objs)
        else:
            self.filled.discard(objs)
    
    def get_filled(self, objs):
        return tuple(obj.name for obj in objs) in self.filled

    def set_saturated(self, objs, is_saturated):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_saturated:
            self.saturated.add(objs)
        else:
            self.saturated.discard(objs)
    
    def get_saturated(self, objs):
        return tuple(obj.name for obj in objs) in self.saturated

    def set_nextto(self, objs, is_nextto):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_nextto:
            self.nextto.add(objs)
        else:
            self.nextto.discard(objs)
    
    def get_nextto(self, objs):
        return tuple(obj.name for obj in objs) in self.nextto

    def set_contains(self, objs, is_contains):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_contains:
            self.contains.add(objs)
        else:
            self.contains.discard(objs)
    
    def get_contains(self, objs):
        return tuple(obj.name for obj in objs) in self.contains
    
    def set_under(self, objs, is_under):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_under:
            self.under.add(objs)
        else:
            self.under.discard(objs)
    
    def get_under(self, objs):
        return tuple(obj.name for obj in objs) in self.under

    def set_touching(self, objs, is_touching):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_touching:
            self.touching.add(objs)
        else:
            self.touching.discard(objs)
    
    def get_touching(self, objs):
        return tuple(obj.name for obj in objs) in self.touching
    
    def set_overlaid(self, objs, is_overlaid):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_overlaid:
            self.overlaid.add(objs)
        else:
            self.overlaid.discard(objs)
    
    def get_overlaid(self, objs):
        return tuple(obj.name for obj in objs) in self.overlaid
    
    def set_attached(self, objs, is_attached):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_attached:
            self.attached.add(objs)
        else:
            self.attached.discard(objs)
    
    def get_attached(self, objs):
        return tuple(obj.name for obj in objs) in self.attached

    def set_draped(self, objs, is_draped):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_draped:
            self.draped.add(objs)
        else:
            self.draped.discard(objs)
    
    def get_draped(self, objs):
        return tuple(obj.name for obj in objs) in self.draped
    
    def set_insource(self, objs, is_insource):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_insource:
            self.insource.add(objs)
        else:
            self.insource.discard(objs)
    
    def get_insource(self, objs):
        return tuple(obj.name for obj in objs) in self.insource
    
    def set_grasped(self, objs, is_grasped):
        assert len(objs) == 2, f"`objs` has length other than 2: {objs}"
        if is_grasped:
            self.grasped.add(objs)
        else:
            self.grasped.discard(objs)
    
    def get_grasped(self, objs):
        return tuple(obj.name for obj in objs) in self.grasped


class TrivialGenericObject(object): 
    def __init__(self, name, simulator):
        self.name = name
        self.simulator = simulator
    
    def get_cooked(self):
        return self.simulator.get_cooked((self,))
    
    def get_frozen(self):
        return self.simulator.get_frozen((self,))
    
    def get_open(self):
        return self.simulator.get_open((self,))
    
    def get_folded(self):
        return self.simulator.get_folded((self,))
    
    def get_unfolded(self):
        return self.simulator.get_unfolded((self,))
    
    def get_toggled_on(self):
        return self.simulator.get_toggled_on((self,))
    
    def get_closed(self):
        return self.simulator.get_closed((self,))
    
    def get_hot(self):
        return self.simulator.get_hot((self,))
    
    def get_on_fire(self):
        return self.simulator.get_on_fire((self,))
    
    def get_empty(self):
        return self.simulator.get_empty((self,))
    
    def get_broken(self):
        return self.simulator.get_broken((self,))
    
    def get_assembled(self):
        return self.simulator.get_assembled((self,))
    
    def get_future(self):
        return self.simulator.get_future((self,))
    
    def get_real(self):
        return self.simulator.get_real((self,))
    
    def get_ontop(self, other):
        return self.simulator.get_ontop((self, other))
    
    def get_covered(self, other):
        return self.simulator.get_covered((self, other))

    def get_inside(self, other):
        return self.simulator.get_inside((self, other))
    
    def get_saturated(self, other):
        return self.simulator.get_saturated((self, other))

    def get_nextto(self, other):
        return self.simulator.get_nextto((self, other))

    def get_contains(self, other):
        return self.simulator.get_contains((self, other))

    def get_under(self, other):
        return self.simulator.get_under((self, other))

    def get_touching(self, other):
        return self.simulator.get_touching((self, other))

    def get_overlaid(self, other):
        return self.simulator.get_overlaid((self, other))

    def get_attached(self, other):
        return self.simulator.get_attached((self, other))

    def get_draped(self, other):
        return self.simulator.get_draped((self, other))

    def get_insource(self, other):
        return self.simulator.get_insource((self, other))
    
    def get_grasped(self, other):
        return self.simulator.get_grasped((self, other))


# OmniGibson trivial predicates
class TrivialCookedPredicate(UnaryAtomicFormula):
    STATE_NAME = "cooked"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_cooked())
        return obj.get_cooked()

    def _sample(self, obj1, binary_state):
        pass


class TrivialFrozenPredicate(UnaryAtomicFormula):
    STATE_NAME = "frozen"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_frozen())
        return obj.get_frozen()

    def _sample(self, obj1, binary_state):
        pass


class TrivialOpenPredicate(UnaryAtomicFormula):
    STATE_NAME = "open"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_open())
        return obj.get_open()

    def _sample(self, obj1, binary_state):
        pass


class TrivialFoldedPredicate(UnaryAtomicFormula):
    STATE_NAME = "folded"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_folded())
        return obj.get_folded()

    def _sample(self, obj1, binary_state):
        pass


class TrivialUnfoldedPredicate(UnaryAtomicFormula):
    STATE_NAME = "unfolded"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_unfolded())
        return obj.get_unfolded()

    def _sample(self, obj1, binary_state):
        pass


class TrivialToggledOnPredicate(UnaryAtomicFormula):
    STATE_NAME = "toggled_on"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_toggled_on())
        return obj.get_toggled_on()

    def _sample(self, obj1, binary_state):
        pass


class TrivialClosedPredicate(UnaryAtomicFormula):
    STATE_NAME = "closed"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_closed())
        return obj.get_closed()

    def _sample(self, obj1, binary_state):
        pass


class TrivialOnFirePredicate(UnaryAtomicFormula):
    STATE_NAME = "on_fire"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_on_fire())
        return obj.get_on_fire()

    def _sample(self, obj1, binary_state):
        pass


class TrivialHotPredicate(UnaryAtomicFormula):
    STATE_NAME = "hot"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_hot())
        return obj.get_hot()

    def _sample(self, obj1, binary_state):
        pass


class TrivialHotPredicate(UnaryAtomicFormula):
    STATE_NAME = "on_fire"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_on_fire())
        return obj.get_on_fire()

    def _sample(self, obj1, binary_state):
        pass


class TrivialEmptyPredicate(UnaryAtomicFormula):
    STATE_NAME = "on_fire"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_empty())
        return obj.get_empty()

    def _sample(self, obj1, binary_state):
        pass


class TrivialBrokenPredicate(UnaryAtomicFormula):
    STATE_NAME = "broken"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_broken())
        return obj.get_broken()

    def _sample(self, obj1, binary_state):
        pass


class TrivialAssembledPredicate(UnaryAtomicFormula):
    STATE_NAME = "assembled"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_assembled())
        return obj.get_assembled()

    def _sample(self, obj1, binary_state):
        pass


class TrivialFuturePredicate(UnaryAtomicFormula):
    STATE_NAME = "future"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_future())
        return obj.get_future()

    def _sample(self, obj1, binary_state):
        pass


class TrivialRealPredicate(UnaryAtomicFormula):
    STATE_NAME = "real"

    def _evaluate(self, obj):
        print(self.STATE_NAME, obj.name, obj.get_real())
        return obj.get_real()

    def _sample(self, obj1, binary_state):
        pass


class TrivialCoveredPredicate(BinaryAtomicFormula):
    STATE_NAME = "covered"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_covered(obj2))
        return obj1.get_covered(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialInsidePredicate(BinaryAtomicFormula):
    STATE_NAME = "inside"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_inside(obj2))
        return obj1.get_inside(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialOnTopPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_ontop(obj2))
        return obj1.get_ontop(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialFilledPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_ontop(obj2))
        return obj1.get_ontop(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialSaturatedPredicate(BinaryAtomicFormula):
    STATE_NAME = "saturated"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_saturated(obj2))
        return obj1.get_saturated(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialNextToPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_nextto(obj2))
        return obj1.get_nextto(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialContainsPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_contains(obj2))
        return obj1.get_contains(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialUnderPredicate(BinaryAtomicFormula):
    STATE_NAME = "under"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_under(obj2))
        return obj1.get_under(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialTouchingPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_touching(obj2))
        return obj1.get_touching(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialOverlaidPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_overlaid(obj2))
        return obj1.get_overlaid(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialAttachedPredicate(BinaryAtomicFormula):
    STATE_NAME = "ontop"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_attached(obj2))
        return obj1.get_attached(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialDrapedPredicate(BinaryAtomicFormula):
    STATE_NAME = "draped"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_draped(obj2))
        return obj1.get_draped(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialInsourcePredicate(BinaryAtomicFormula):
    STATE_NAME = "insource"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_insource(obj2))
        return obj1.get_insource(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


class TrivialGraspedPredicate(BinaryAtomicFormula):
    STATE_NAME = "grasped"

    def _evaluate(self, obj1, obj2):
        print(self.STATE_NAME, obj1.name, obj2.name, obj1.get_grasped(obj2))
        return obj1.get_grasped(obj2)

    def _sample(self, obj1, obj2, binary_state):
        pass


VALID_ATTACHMENTS = set([
    ("mixing_bowl.n.01", "electric_mixer.n.01"),
    ("cork.n.04", "wine_bottle.n.01"),
    ("menu.n.01", "wall.n.01"),
    ("broken__light_bulb.n.01", "table_lamp.n.01"),
    ("light_bulb.n.01", "table_lamp.n.01"),
    ("lens.n.01", "digital_camera.n.01"),
    ("screen.n.01", "wall.n.01"),
    ("antler.n.01", "wall.n.01"),
    ("skateboard_wheel.n.01", "skateboard.n.01"),
    ("blackberry.n.01", "scrub.n.01"),
    ("raspberry.n.02", "scrub.n.01"),
    ("dip.n.07", "candlestick.n.01"),
    ("sign.n.02", "wall.n.01"),
    ("wreath.n.01", "wall.n.01"),
    ("bow.n.08", "wall.n.01"),
    ("holly.n.03", "wall.n.01"),
    ("curtain_rod.n.01", "wall.n.01"),
    ("bicycle.n.01", "bicycle_rack.n.01"),
    ("bicycle_rack.n.01", "wall.n.01"),
    ("dartboard.n.01", "wall.n.01"),
    ("rug.n.01", "wall.n.01"),
    ("fairy_light.n.01", "wall.n.01"),
    ("lantern.n.01", "wall.n.01"),
    ("address.n.05", "wall.n.01"),
    ("hanger.n.02", "wardrobe.n.01"),
    ("flagpole.n.02", "wall.n.01"),
    ("picture_frame.n.01", "wall.n.01"),
    ("wind_chime.n.01", "pole.n.01"),
    ("pole.n.01", "wall.n.01"),
    ("hook.n.05", "trailer_truck.n.01"),
    ("fire_alarm.n.02", "wall.n.01"),
    ("poster.n.01", "wall.n.01"),
    ("painting.n.01", "wall.n.01"),
    ("hanger.n.02", "coatrack.n.01"),
    ("license_plate.n.01", "car.n.01"),
    ("gummed_label.n.01", "license_plate.n.01"),
    ("wallpaper.n.01", "wall.n.01"),
    ("mirror.n.01", "wall.n.01"),
    ("webcam.n.02", "desktop_computer.n.01"),
    ("kayak.n.01", "kayak_rack.n.01"),
    ("kayak_rack.n.01", "wall.n.01"),
    ("fish.n.02", "fishing_rod.n.01"),
    ("bicycle_rack.n.01", "recreational_vehicle.n.01"),
])

VALID_ROOMS = set()


import json
import os 
import pandas as pd
import pathlib

from bddl.data_generation.process_prop_param_annots import LEAF_SYNSETS_SET


TRANSITION_MAP_DIR = pathlib.Path(__file__).parents[1] / "generated_data" / "transition_map" / "tm_jsons"


def generate_slicing_rules(syns_to_param_props, props_to_syns):
    """
    Generates transition rules for slicing of sliceable objects.
    Form: 
    {
        "rule_name": <non-unique string>,
        "input_synsets": {
            <sliceable>: 1 
        },
        "output_synsets": {
            <sliceable_derivative_synset>: 2
        }
    }
    Assumptions not listed in rule: 
    - Slicer is present and slicing region of slicer is touching sliceable - this is required for the rule to fire
    """

    rules = []
    sliceables = props_to_syns["sliceable"]
    for sliceable in sliceables: 
        assert "sliceable_derivative_synset" in syns_to_param_props[sliceable]["sliceable"], f"Synset {sliceable} has no sliceable derivative synset. Add one in the parameter annotations or handle otherwise."
        assert syns_to_param_props[sliceable]["sliceable"]["sliceable_derivative_synset"] in syns_to_param_props, f"Synset {sliceable} has sliceable derivative synset {syns_to_param_props[sliceable]['sliceable_derivative_synset']} that is not a valid synset. Please check."
        sliceable_derivative_synset = syns_to_param_props[sliceable]["sliceable"]["sliceable_derivative_synset"]

        rule = {
            "rule_name": f"{sliceable}-slicing",
            "input_synsets": {
                sliceable: 1
            },
            "output_synsets": {
                sliceable_derivative_synset: 2
            }
        }

        rules.append(rule)
    
    rules.sort(key=lambda x: x["rule_name"])
    return rules


def generate_dicing_rules(syns_to_param_props, props_to_syns):
    """
    Generates transition rules for dicing of diceable objects.
    Form: 
    {
        "rule_name": <unique string name>,
        "input_synsets": {
            <diceable>: 1 
        },
        "input_states": {
            <diceable>: [
                [
                    "cooked",
                    <bool>
                ]
            ]
        }
        "output_synsets": {
            <cooked_diceable_derivative_synset> if cooked else <uncooked_diceable_derivative_synset>: 1
        }
    }
    Assumptions not listed in rule: 
    - Slicer is present and slicing region of slicer is touching diceable - this is required for the rule to fire
    """

    rules = []
    diceables = props_to_syns["diceable"]
    cookables = set(props_to_syns["cookable"])
    for diceable in diceables: 
        assert "uncooked_diceable_derivative_synset" in syns_to_param_props[diceable]["diceable"], f"Synset {diceable} has no uncooked diceable derivative synset. Add one in the parameter annotations."
        assert syns_to_param_props[diceable]["diceable"]["uncooked_diceable_derivative_synset"] in syns_to_param_props, f"Synset {diceable} has uncooked diceable derivative synset {syns_to_param_props[diceable]['uncooked_diceable_derivative_synset']} that is not a valid synset. Please check."
        uncooked_diceable_derivative_synset = syns_to_param_props[diceable]["diceable"]["uncooked_diceable_derivative_synset"]

        uncooked_rule = {
            "rule_name": f"uncooked-{diceable}-dicing",
            "input_synsets": {
                diceable: 1
            },
            "input_states": {
                diceable: [
                    [
                        "cooked",
                        False
                    ]
                ]
            },
            "output_synsets": {
                uncooked_diceable_derivative_synset: 1
            }
        }
        rules.append(uncooked_rule)

        if diceable in cookables:
            assert "cooked_diceable_derivative_synset" in syns_to_param_props[diceable]["diceable"], f"Synset {diceable} has no cooked diceable derivative synset. Add one in the parameter annotations."
            assert syns_to_param_props[diceable]["diceable"]["cooked_diceable_derivative_synset"] in syns_to_param_props, f"Synset {diceable} has cooked diceable derivative synset {syns_to_param_props[diceable]['cooked_diceable_derivative_synset']} that is not a valid synset. Please check."
            cooked_diceable_derivative_synset = syns_to_param_props[diceable]["diceable"]["cooked_diceable_derivative_synset"]

            cooked_rule = {
                "rule_name": f"cooked-{diceable}-dicing",
                "input_synsets": {
                    diceable: 1
                },
                "input_states": {
                    diceable: [
                        [
                            "cooked",
                            True
                        ]
                    ]
                },
                "output_synsets": {
                    cooked_diceable_derivative_synset: 1
                }
            }
            rules.append(cooked_rule)
    
    rules.sort(key=lambda x: x["rule_name"])
    return rules 


def generate_substance_cooking_rules(syns_to_param_props, props_to_syns):
    """
    Generates transition rules for dicing of diceable objects.
    Form: 
    {
        "rule_name": <non-unique string>,
        "input_synsets": {
            <cookable_substance>: 1
        },
        "output_synsets": {
            <substance_cooking_derivative_synset>: 1
        }
    }
    Assumptions not listed in rule: 
    TODO is the below accurate? 
    - Cookable substance must be filling a fillable and the fillable must be hot for the rule to fire
    """

    rules = []
    cookable_substances = set(props_to_syns["cookable"]).intersection(set(props_to_syns["substance"])).difference(set(props_to_syns["waterCook"]))
    for cookable_substance in cookable_substances: 
        assert "substance_cooking_derivative_synset" in syns_to_param_props[cookable_substance]["cookable"], f"Synset {cookable_substance} has no substance cooking derivative synset. Add one in the parameter annotations or handle otherwise."
        assert syns_to_param_props[cookable_substance]["cookable"]["substance_cooking_derivative_synset"] in syns_to_param_props, f"Synset {cookable_substance} has substance cooking derivative synset {syns_to_param_props[cookable_substance]['substance_cooking_derivative_synset']} that is not a valid synset. Please check."
        substance_cooking_derivative_synset = syns_to_param_props[cookable_substance]["cookable"]["substance_cooking_derivative_synset"]

        rule = {
            "rule_name": f"{cookable_substance}-cooking",
            "input_synsets": {
                cookable_substance: 1
            },
            "output_synsets": {
                substance_cooking_derivative_synset: 1
            }
        }
        rules.append(rule)
    
    rules.sort(key=lambda x: x["rule_name"])
    return rules 


def generate_substance_watercooking_rules(syns_to_param_props, props_to_syns):
    """
    Generates transition rules for dicing of diceable objects.
    Form: 
    {
        "rule_name": <non-unique string>,
        "input_synsets": {
            <watercookable_substance>: 1,
            water.n.06_1: 1
        },
        "output_synsets": {
            <substance_cooking_derivative_synset>: 1
        }
    }
    Assumptions not listed in rule: 
    TODO is the below accurate? 
    - Cookable substance must be filling a fillable and the fillable must be hot for the rule to fire
    """

    rules = []
    watercookable_substances = set(props_to_syns["waterCook"])
    for watercookable_substance in watercookable_substances: 
        assert "cookable" in syns_to_param_props[watercookable_substance], f"Synset {watercookable_substance} has no cookable property annotation."
        assert "substance_cooking_derivative_synset" in syns_to_param_props[watercookable_substance]["cookable"], f"Synset {watercookable_substance} has no substance cooking derivative synset. Add one in the parameter annotations or handle otherwise."
        assert syns_to_param_props[watercookable_substance]["cookable"]["substance_cooking_derivative_synset"] in syns_to_param_props, f"Synset {watercookable_substance} has substance cooking derivative synset {syns_to_param_props[watercookable_substance]['substance_cooking_derivative_synset']} that is not a valid synset. Please check."
        substance_cooking_derivative_synset = syns_to_param_props[watercookable_substance]["cookable"]["substance_cooking_derivative_synset"]

        rule = {
            "rule_name": f"{watercookable_substance}-cooking",
            "input_synsets": {
                watercookable_substance: 1,
                "cooked__water.n.01": 1
            },
            "output_synsets": {
                substance_cooking_derivative_synset: 1
            }
        }
        rules.append(rule)
    
    rules.sort(key=lambda x: x["rule_name"])
    return rules 


def generate_melting_rules(syns_to_param_props, props_to_syns):
    """
    Generates transition rules for melting of meltable objects.
    Form: 
    {
        "rule_name": <unique string name>,
        "input_synsets": {
            <meltable>: 1
        },
        "output_synsets": {
            <meltable_derivative_synset>: 1
        }
    }
    """
    rules = []
    meltables = props_to_syns["meltable"]
    for meltable in meltables: 
        assert "meltable_derivative_synset" in syns_to_param_props[meltable]["meltable"], f"Synset {meltable} has no meltable derivative synset. Add one in the parameter annotations or handle otherwise."
        assert syns_to_param_props[meltable]["meltable"]["meltable_derivative_synset"] in syns_to_param_props, f"Synset {meltable} has meltable derivative synset {syns_to_param_props[meltable]['meltable_derivative_synset']} that is not a valid synset. Please check."
        meltable_derivative_synset = syns_to_param_props[meltable]["meltable"]["meltable_derivative_synset"]

        rule = {
            "rule_name": f"{meltable}-melting",
            "input_synsets": {
                meltable: 1
            },
            "output_synsets": {
                meltable_derivative_synset: 1
            }
        }

        rules.append(rule)
    
    rules.sort(key=lambda x: x["rule_name"])
    return rules


def generate_washer_particleremover_rules(props_to_syns):
    """
    Generates transition rules for adding water to particleRemovers that come out of a washer
    Form: 
    {
        "rule_name": <particleRemover>-washer-saturate-cover,
        "input_synsets": {},
        "washed_item": {
            <particleRemover>: 1,
        },
        "output_synsets": {
            "water.n.06": 1
        }
    }
    Invariants not explicitly stated in rule: 
    - At input time washer.n.03 exists and is toggled on, `washed_item` is inside
    - At output time `washed_item` is covered and saturated in water
    """
    rules = []
    particleRemovers = set(props_to_syns["particleRemover"])
    for particleRemover in particleRemovers: 
        rule = {
            "rule_name": f"{particleRemover}-washer-saturate-cover",
            "input_synsets": {},
            "washed_item": {
                particleRemover: 1
            },
            "output_synsets": {
                "water.n.06": 1
            }
        }
        rules.append(rule)
    rules.sort(key=lambda x: x["rule_name"])
    return rules


def generate_washer_nonparticleremover_rules(props_to_syns):
    """
    Generates transition rules for adding water to non-particleRemover nonSubstances that come 
    out of a washer.
    Form: 
    {
        "rule_name": <nonSubstance,non-particleRemover>-washer-cover,
        "input_synsets": {},
        "washed_item": {
            <non-particleRemover,nonSubstance>: 1
        },
        "output_synsets": {
            "water.n.06": 1
        }
    }
    Invariants not explicitly stated in rule: 
    - At input time washer.n.03 exists and is toggled on, `washed_item` is
    inside
    - At output time `washed_item` is covered and saturated in water
    """
    rules = []
    particleRemovers = set(props_to_syns["particleRemover"])
    nonSubstances = set(props_to_syns["nonSubstance"])
    nonparticleremover_nonsubstances = nonSubstances.difference(particleRemovers)
    for syn in nonparticleremover_nonsubstances: 
        rule = {
            "rule_name": f"{syn}-washer-cover",
            "input_synsets": {},
            "washed_item": {
                syn: 1
            },
            "output_synsets": {
                "water.n.06": 1
            }
        }
        rules.append(rule)
    rules.sort(key=lambda x: x["rule_name"])
    return rules


def create_get_save_implicit_transition_rules(syns_to_param_props, props_to_syns):
    print("Creating implicit transition rule jsons...")
    # Constrain to leaf synsets
    for prop in props_to_syns:
        props_to_syns[prop] = set(props_to_syns[prop]) & LEAF_SYNSETS_SET
    slicing_rules = generate_slicing_rules(syns_to_param_props, props_to_syns)
    dicing_rules = generate_dicing_rules(syns_to_param_props, props_to_syns)
    substance_cooking_rules = generate_substance_cooking_rules(syns_to_param_props, props_to_syns)
    substance_watercooking_rules = generate_substance_watercooking_rules(syns_to_param_props, props_to_syns)
    melting_rules = generate_melting_rules(syns_to_param_props, props_to_syns)
    washer_particleremover_rules = generate_washer_particleremover_rules(props_to_syns)
    washer_nonparticleremover_rules = generate_washer_nonparticleremover_rules(props_to_syns)

    with open(os.path.join(TRANSITION_MAP_DIR, "slicing.json"), "w") as f:
        json.dump(slicing_rules, f, indent=4)
    with open(os.path.join(TRANSITION_MAP_DIR, "dicing.json"), "w") as f:
        json.dump(dicing_rules, f, indent=4)
    with open(os.path.join(TRANSITION_MAP_DIR, "substance_cooking.json"), "w") as f:
        json.dump(substance_cooking_rules, f, indent=4)
    with open(os.path.join(TRANSITION_MAP_DIR, "substance_watercooking.json"), "w") as f:
        json.dump(substance_watercooking_rules, f, indent=4) 
    with open(os.path.join(TRANSITION_MAP_DIR, "melting.json"), "w") as f:
        json.dump(melting_rules, f, indent=4)
    with open(os.path.join(TRANSITION_MAP_DIR, "washer_particleremover.json"), "w") as f:
        json.dump(washer_particleremover_rules, f, indent=4)
    with open(os.path.join(TRANSITION_MAP_DIR, "washer_nonparticleremover.json"), "w") as f:
        json.dump(washer_nonparticleremover_rules, f, indent=4)
    print("Created and saved implicit transition rule jsons.")